==> diablo-1.05.tgz: diablo/lib/history.c <==

/*
 * LIB/HISTORY.C
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"

Prototype void HistoryOpen(const char *fileName, int fastMode);
Prototype int  HistoryClose(void);
Prototype int HistoryLookup(const char *msgid, FILE **pfi);
Prototype int HistoryLookupByHash(hash_t hv, History *h);
Prototype int HistoryAdd(const char *msgid, History *h);
Prototype hash_t hhash(const char *msgid);
Prototype char *MsgId(char *s);

#define P1	7834891
#define P2	6017489

unsigned long	*HAry;
int		HFd = -1;
int		LoggedDHistCorrupt;
int		HFlags;
int		HSize;
int		HMask;

void
HistoryOpen(const char *fileName, int hflags)
{
    int fd;
    struct stat st;

    HFlags = hflags;

    /*
     * open the history file.  Require a dhistory.param file if
     * the hash table size is not 1024*1024.
     */

    if (fileName)
	fd = xopen(O_RDWR|O_CREAT, 0644, "%s", fileName);
    else
	fd = xopen(O_RDWR|O_CREAT, 0644, "%s/dhistory", NewsHome);

    if (fd < 0) {
	syslog(LOG_ERR, "open %s/dhistory failed", NewsHome);
	exit(1);
    }

    if (fstat(fd, &st) < 0) {
	syslog(LOG_ERR, "fstat %s/dhistory failed", NewsHome);
	exit(1);
    }

    /*
     * open/check the parameter file
     */

    if (st.st_size != 0) {
	FILE *fi;

	if (fileName)
	    fi = xfopen("r", "%s.param", fileName);
	else
	    fi = xfopen("r", "%s/dhistory.param", NewsHome);

	HSize = 0;

	if (fi) {
	    char buf[256];

	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		if (buf[0] == 'H')
		    HSize = strtol(buf + 1, NULL, 0);
	    }
	    fclose(fi);

	    if (HSize == 0) {
		syslog(LOG_ERR, "dhistory parameter file format error");
		fprintf(stderr, "dhistory parameter file format error\n");
		exit(1);
	    }
	}

	if (HSize == 0)
	    HSize = 1024 * 1024;

	HMask = HSize - 1;
    }

    /*
     * initial history file creation, if necessary
     */

    if (st.st_size == 0) {
	char *z = calloc(4096, 1);

	xflock(fd, XLOCK_EX);

	fstat(fd, &st);

	if (st.st_size == 0) {
	    /*
	     * create history param file
	     */
	    {
		int fdp;
		char buf[64];

		if (fileName)
		    fdp = xopen(O_RDWR|O_CREAT, 0644, "%s.param", fileName);
		else
		    fdp = xopen(O_RDWR|O_CREAT, 0644, "%s/dhistory.param", NewsHome);
		if (fdp < 0) {
		    syslog(LOG_ERR, "cannot create .param file");
		    exit(1);
		}
		sprintf(buf, "H%d\n", HSIZE);
		write(fdp, buf, strlen(buf));
		close(fdp);
	    }

	    HSize = HSIZE;
	    HMask = HSize - 1;

	    lseek(fd, st.st_size, 0);
	    while (st.st_size < HSIZE * sizeof(long)) {
		long n = HSIZE * sizeof(long) - st.st_size;
		if (n > 4096)
		    n = 4096;
		if (write(fd, z, n) != n) {
		    syslog(LOG_ERR, "dhistory hash table creation failed");
		    exit(1);
		}
		st.st_size += n;
	    }
	}
	free(z);

	xflock(fd, XLOCK_UN);
    }

    /*
     * In FAST mode we leave the history file locked in order to
     * cache the hash table array at the beginning, which in turn
     * allows us to essentially append new entries to the end of
     * the file without having to seek back and forth updating
     * the hash table.
     *
     * When we aren't in FAST mode, we memory-map the hash table
     * portion of the history file.
     */

    if (HFlags & HGF_FAST)
	xflock(fd, XLOCK_EX);

    if (HFlags & HGF_FAST) {
	HAry = calloc(HSize, sizeof(long));
	if (HAry == NULL) {
	    perror("calloc");
	    exit(1);
	}
	lseek(fd, 0L, 0);
	if (read(fd, HAry, HSize * sizeof(long)) != HSize * sizeof(long)) {
	    perror("read");
	    exit(1);
	}
    } else {
	HAry = (void *)mmap((caddr_t)0, HSize * sizeof(long), PROT_READ, MAP_SHARED, fd, 0);
    }

    if (HAry == NULL || HAry == (unsigned long *)-1) {
	if (fd >= 0)
	    close(fd);
	perror("mmap");
	exit(1);
    }
    HFd = fd;
}

/*
 * On close, we have to commit the hash table if we were in
 * FAST mode, otherwise we need only unmap the file before
 * closing it.
 */

int
HistoryClose(void)
{
    int r = RCOK;

    if (HFd >= 0) {
	if (HFlags & HGF_FAST) {
	    lseek(HFd, 0L, 0);
	    if (write(HFd, HAry, HSize * sizeof(long)) != HSize * sizeof(long))
		r = RCTRYAGAIN;
	    else
		free(HAry);
	} else {
	    if (HAry && HAry != (unsigned long *)-1)
		munmap((void *)HAry, HSize * sizeof(long));
	}
    }
    if (r == RCOK) {
	if (HFd >= 0)
	    close(HFd);
	HFd = -1;
	HAry = NULL;
    }
    return(r);
}

int
HistoryLookup(const char *msgid, FILE **pfi)
{
    hash_t hv = hhash(msgid);
    long poff = ((hv.h1 ^ hv.h2) & HMask) * sizeof(long);
    long off = HAry[(hv.h1 ^ hv.h2) & HMask];
    History h = { 0 };

    while (off) {
	lseek(HFd, off, 0);
	if (read(HFd, &h, sizeof(h)) != sizeof(h)) {
	    if ((LoggedDHistCorrupt & 1) == 0 || DebugOpt) {
		LoggedDHistCorrupt |= 1;
		syslog(LOG_ERR, "dhistory file corrupted on lookup");
		sleep(1);
	    }
	    break;
	}
	if (h.hv.h1 == hv.h1 && h.hv.h2 == hv.h2)
	    break;
	poff = off;
	off = h.next;
    }
    if (off != 0) {
	if (pfi) {
	    *pfi = xfopen("r", "D.%08lx/%08lx.%08lx.%04x", 
		h.gmt - h.gmt % 10, 
		hv.h1, 
		hv.h2, 
		h.iter
	    );
	}
	return(0);
    }
    return(-1);
}

int
HistoryLookupByHash(hash_t hv, History *h)
{
    long poff = ((hv.h1 ^ hv.h2) & HMask) * sizeof(long);
    long off = HAry[(hv.h1 ^ hv.h2) & HMask];

    while (off) {
	lseek(HFd, off, 0);
	if (read(HFd, h, sizeof(*h)) != sizeof(*h)) {
	    if ((LoggedDHistCorrupt & 1) == 0 || DebugOpt) {
		LoggedDHistCorrupt |= 1;
		syslog(LOG_ERR, "dhistory file corrupted on lookup");
		sleep(1);
	    }
	    break;
	}
	if (h->hv.h1 == hv.h1 && h->hv.h2 == hv.h2)
	    break;
	poff = off;
	off = h->next;
    }
    if (off != 0) {
	return(0);
    }
    return(-1);
}

int
HistoryAdd(const char *msgid, History *h)
{
    long hi = (h->hv.h1 ^ h->hv.h2) & HMask;
    long poff = hi * sizeof(long);
    long boff = poff;
    long off;
    int r = RCOK;

    if ((HFlags & HGF_FAST) == 0)
	hflock(HFd, boff, XLOCK_EX);

    off = HAry[hi];

    if ((HFlags & HGF_NOSEARCH) == 0) {
	while (off) {
	    History ht;

	    lseek(HFd, off, 0);
	    if (read(HFd, &ht, sizeof(ht)) != sizeof(ht)) {
		if ((LoggedDHistCorrupt & 2) == 0 || DebugOpt) {
		    LoggedDHistCorrupt |= 2;
		    syslog(LOG_ERR, "dhistory file corrupted @ %d", off);
		    sleep(1);
		}
		break;
	    }
	    if (ht.hv.h1 == h->hv.h1 && ht.hv.h2 == h->hv.h2) {
		r = RCALREADY;
		break;
	    }
	    poff = off;
	    off = ht.next;
	}
    }

    if (r == RCOK) {
	fcntl(HFd, F_SETFL, O_APPEND);

	/*
	 * We insert the record into the chain at the beginning of the
	 * chain, even though we are appending the record.  This is for
	 * time-locality of reference and also reduces 'append-to-history'
	 * file overhead by localizing disk writes a bit better.
	 */

	h->next = HAry[hi];

	if (write(HFd, h, sizeof(History)) == sizeof(History)) {
	    off = lseek(HFd, 0L, 1) - sizeof(History);

	    fcntl(HFd, F_SETFL, 0);
	    if ((HFlags & HGF_FAST) == 0) {
		lseek(HFd, hi * sizeof(long), 0);
		if (write(HFd, &off, sizeof(long)) != sizeof(long)) {
		    ftruncate(HFd, off);
		    r = RCTRYAGAIN;
		}
	    } else {
		HAry[hi] = off;
	    }
	} else {
	    fcntl(HFd, F_SETFL, 0);
	    r = RCTRYAGAIN;
	}
    }

    if ((HFlags & HGF_FAST) == 0)
	hflock(HFd, boff, XLOCK_UN);

    return(r);
}

hash_t
hhash(const char *msgid)
{
    int h1 = 0x0034629D;
    int h2 = 0x007395DD;
    int hv = 0x1A3F5C4F;
    hash_t t;

    while (*msgid) {
	h1 = h1 * *(const unsigned char *)msgid % P1;
	h2 = h2 * *(const unsigned char *)msgid % P2;
	hv = (hv << 5) ^ *(const unsigned char *)msgid ^ (hv >> 23);
	++msgid;
    }
    t.h1 = (h1 ^ (hv << 14)) & 0x7FFFFFFF;	/* bit 31 reserved for future use */
    t.h2 = (h2 ^ (hv << 20)) & 0x7FFFFFFF;	/* bit 31 reserved for future use */
    return(t);
}

char *
MsgId(char *s)
{
    int i;
    static char *LMsgId;

    if (LMsgId) {
	free(LMsgId);
	LMsgId = NULL;
    }

    if (s == NULL)
	return("<>");
    while (*s && (*s == ' ' || *s == '\t'))
	++s;
    if (*s != '<')
	return("<>");
    for (i = 0; s[i] && s[i] != '>'; ++i)
	;
    if (s[i] != '>')
	return("<>");
    if (i > MAXMSGIDLEN)
	return("<>");
    s[i+1] = 0;
    LMsgId = strcpy(malloc(strlen(s) + 1), s);
    return(LMsgId);
}

==> diablo-1.08.tgz: diablo/lib/history.c <==

/*
 * LIB/HISTORY.C
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"

Prototype void HistoryOpen(const char *fileName, int fastMode);
Prototype int  HistoryClose(void);
Prototype int HistoryLookup(const char *msgid, char **pfi, long *rsize, int *pmart);
Prototype int HistoryLookupByHash(hash_t hv, History *h);
Prototype int HistoryAdd(const char *msgid, History *h);
Prototype hash_t hhash(const char *msgid);
Prototype char *MsgId(char *s);

#define P1	7834891
#define P2	6017489

HistHead	HHead;
unsigned long	*HAry;
int		HFd = -1;
int		LoggedDHistCorrupt;
int		HFlags;
int		HSize;
int		HMask;

void
HistoryOpen(const char *fileName, int hflags)
{
    int fd;
    struct stat st;

    HFlags = hflags;

    /*
     * open the history file
     */

    if (fileName)
	fd = xopen(O_RDWR|O_CREAT, 0644, "%s", fileName);
    else
	fd = xopen(O_RDWR|O_CREAT, 0644, "%s/dhistory", NewsHome);

    if (fd < 0) {
	syslog(LOG_ERR, "open %s/dhistory failed", NewsHome);
	exit(1);
    }

    if (fstat(fd, &st) < 0) {
	syslog(LOG_ERR, "fstat %s/dhistory failed", NewsHome);
	exit(1);
    }

    /*
     * initial history file creation, if necessary
     */

    bzero(&HHead, sizeof(HHead));

    if (st.st_size == 0 || 
	read(fd, &HHead, sizeof(HHead)) != sizeof(HHead) ||
	HHead.hmagic != HMAGIC
    ) {
	/*
	 * lock after finding the history file to be invalid and recheck.
	 */

	xflock(fd, XLOCK_EX);

	fstat(fd, &st);
	lseek(fd, 0L, 0);

	if (st.st_size == 0 || 
	    read(fd, &HHead, sizeof(HHead)) != sizeof(HHead) ||
	    HHead.hmagic != HMAGIC
	) {
	    int n;
	    int b;
	    char *z = calloc(4096, 1);

	    /*
	     * check for old version of history file
	     */

	    if (st.st_size) {
		syslog(LOG_ERR, "Incompatible history file version or corrupted history file\n");
		exit(1);
	    }

	    /*
	     * create new history file
	     */

	    lseek(fd, 0L, 0);
	    ftruncate(fd, 0);
	    bzero(&HHead, sizeof(HHead));

	    HHead.hashSize = HSIZE;
	    HHead.version  = HVERSION;
	    HHead.henSize  = sizeof(History);
	    HHead.headSize = sizeof(HHead);

	    write(fd, &HHead, sizeof(HHead));

	    /*
	     * write out the hash table
	     */

	    n = 0;
	    b = HHead.hashSize * sizeof(long);

	    while (n < b) {
		int r = (b - n > 4096) ? 4096 : b - n;

		write(fd, z, r);
		n += r;
	    }
	    fsync(fd);

	    /*
	     * rewrite header with magic number
	     */

	    lseek(fd, 0L, 0);
	    HHead.hmagic = HMAGIC;
	    write(fd, &HHead, sizeof(HHead));

	    free(z);
	}

	xflock(fd, XLOCK_UN);
    }

    if (HHead.version != HVERSION) {
	syslog(LOG_ERR, "DHistory version %d, expected %d\n",
	    HHead.version,
	    HVERSION
	);
	exit(1);
    }

    /*
     * Map history file
     */

    HSize = HHead.hashSize;
    HMask = HSize - 1;

    /*
     * In FAST mode we leave the history file locked in order to
     * cache the hash table array at the beginning, which in turn
     * allows us to essentially append new entries to the end of
     * the file without having to seek back and forth updating
     * the hash table.
     *
     * When we aren't in FAST mode, we memory-map the hash table
     * portion of the history file.
     */

    if (HFlags & HGF_FAST)
	xflock(fd, XLOCK_EX);

    if (HFlags & HGF_FAST) {
	HAry = calloc(HSize, sizeof(long));
	if (HAry == NULL) {
	    perror("calloc");
	    exit(1);
	}
	lseek(fd, HHead.headSize, 0);
	if (read(fd, HAry, HSize * sizeof(long)) != HSize * sizeof(long)) {
	    perror("read");
	    exit(1);
	}
    } else {
	HAry = xmap(NULL, HSize * sizeof(long), PROT_READ, MAP_SHARED, fd, HHead.headSize);
    }

    if (HAry == NULL || HAry == (unsigned long *)-1) {
	if (fd >= 0)
	    close(fd);
	perror("mmap");
	exit(1);
    }
    HFd = fd;
}

/*
 * On close, we have to commit the hash table if we were in
 * FAST mode, otherwise we need only unmap the file before
 * closing it.
 */

int
HistoryClose(void)
{
    int r = RCOK;

    if (HFd >= 0) {
	if (HFlags & HGF_FAST) {
	    lseek(HFd, HHead.headSize, 0);
	    if (write(HFd, HAry, HSize * sizeof(long)) != HSize * sizeof(long))
		r = RCTRYAGAIN;
	    else
		free(HAry);
	} else {
	    if (HAry && HAry != (unsigned long *)-1)
		xunmap((void *)HAry, HSize * sizeof(long));
	}
    }
    if (r == RCOK) {
	if (HFd >= 0)
	    close(HFd);
	HFd = -1;
	HAry = NULL;
    }
    return(r);
}

int
HistoryLookup(const char *msgid, char **pfi, long *rsize, int *pmart)
{
    hash_t hv = hhash(msgid);
    long poff = HHead.headSize + ((hv.h1 ^ hv.h2) & HMask) * sizeof(long);
    long off = HAry[(hv.h1 ^ hv.h2) & HMask];
    History h = { 0 };

    while (off) {
	lseek(HFd, off, 0);
	if (read(HFd, &h, sizeof(h)) != sizeof(h)) {
	    if ((LoggedDHistCorrupt & 1) == 0 || DebugOpt) {
		LoggedDHistCorrupt |= 1;
		syslog(LOG_ERR, "dhistory file corrupted on lookup");
		sleep(1);
	    }
	    break;
	}
	if (h.hv.h1 == hv.h1 && h.hv.h2 == hv.h2)
	    break;
	poff = off;
	off = h.next;
    }
    if (off != 0) {
	if (pfi) {
	    char path[128];
	    int fd;

	    ArticleFileName(path, &h, 1);

	    /*
	     * multi-article file ?  If so, articles are zero-terminated
	     */

	    *pmart = 0;
	    if (h.boffset || h.bsize)
		*pmart = 1;

	    /*
	     * get the file
	     */

	    *rsize = 0;
	    if ((fd = cdopen(path, O_RDONLY, 0)) >= 0) {
		*rsize = h.bsize;
		*pfi = xmap(NULL, h.bsize + *pmart, PROT_READ, MAP_SHARED, fd, h.boffset);
		if (*pfi == (char *)-1)
		    *pfi = NULL;
	    }
	    close(fd);
	}
	return(0);
    }
    return(-1);
}

int
HistoryLookupByHash(hash_t hv, History *h)
{
    long poff = HHead.headSize + ((hv.h1 ^ hv.h2) & HMask) * sizeof(long);
    long off = HAry[(hv.h1 ^ hv.h2) & HMask];

    while (off) {
	lseek(HFd, off, 0);
	if (read(HFd, h, sizeof(*h)) != sizeof(*h)) {
	    if ((LoggedDHistCorrupt & 1) == 0 || DebugOpt) {
		LoggedDHistCorrupt |= 1;
		syslog(LOG_ERR, "dhistory file corrupted on lookup");
		sleep(1);
	    }
	    break;
	}
	if (h->hv.h1 == hv.h1 && h->hv.h2 == hv.h2)
	    break;
	poff = off;
	off = h->next;
    }
    if (off != 0) {
	return(0);
    }
    return(-1);
}

int
HistoryAdd(const char *msgid, History *h)
{
    long hi = (h->hv.h1 ^ h->hv.h2) & HMask;
    long poff = HHead.headSize + hi * sizeof(long);
    long boff = poff;
    long off;
    int r = RCOK;

    if ((HFlags & HGF_FAST) == 0)
	hflock(HFd, boff, XLOCK_EX);

    off = HAry[hi];

    if ((HFlags & HGF_NOSEARCH) == 0) {
	while (off) {
	    History ht;

	    lseek(HFd, off, 0);
	    if (read(HFd, &ht, sizeof(ht)) != sizeof(ht)) {
		if ((LoggedDHistCorrupt & 2) == 0 || DebugOpt) {
		    LoggedDHistCorrupt |= 2;
		    syslog(LOG_ERR, "dhistory file corrupted @ %d", off);
		    sleep(1);
		}
		break;
	    }
	    if (ht.hv.h1 == h->hv.h1 && ht.hv.h2 == h->hv.h2) {
		r = RCALREADY;
		break;
	    }
	    poff = off;
	    off = ht.next;
	}
    }

    if (r == RCOK) {
	/*
	 * exclusive lock for append.  It turns out that a number of
	 * kernels, including FreeBSD, do not guarentee atomic appends
	 * under very heavy loads so we serialize history file appends...
	 * hopefully this will be mostly cached, considering how small
	 * the write is and locality of reference.
	 */

	hflock(HFd, 0, XLOCK_EX);

	/*
	 * check alignment before write
	 */

	off = lseek(HFd, 0L, 2) - HHead.headSize - HHead.hashSize*sizeof(long);
	off = off % HHead.henSize;

	if (off) {
	    char cbuf[sizeof(History)] = { 0 };
	    write(HFd, cbuf, HHead.henSize - off);
	    syslog(LOG_ERR, "dhistory file realigned by %d @%d\n", HHead.henSize - off, lseek(HFd, 0L, 1));
	}

	/*
	 * We insert the record into the chain at the beginning of the
	 * chain, even though we are appending the record.  This is for
	 * time-locality of reference and also reduces 'append-to-history'
	 * file overhead by localizing disk writes a bit better.
	 */

	h->next = HAry[hi];

	if (write(HFd, h, sizeof(History)) == sizeof(History)) {
	    off = lseek(HFd, 0L, 1) - sizeof(History);

	    if ((HFlags & HGF_FAST) == 0) {
		lseek(HFd, HHead.headSize + hi * sizeof(long), 0);
		if (write(HFd, &off, sizeof(long)) != sizeof(long)) {
		    ftruncate(HFd, off);
		    r = RCTRYAGAIN;
		}
	    } else {
		HAry[hi] = off;
	    }
	} else {
	    r = RCTRYAGAIN;
	}
	hflock(HFd, 0, XLOCK_UN);
    }

    if ((HFlags & HGF_FAST) == 0)
	hflock(HFd, boff, XLOCK_UN);

    return(r);
}

hash_t
hhash(const char *msgid)
{
    int h1 = 0x0034629D;
    int h2 = 0x007395DD;
    int hv = 0x1A3F5C4F;
    hash_t t;

    while (*msgid) {
	h1 = h1 * *(const unsigned char *)msgid % P1;
	h2 = h2 * *(const unsigned char *)msgid % P2;
	hv = (hv << 5) ^ *(const unsigned char *)msgid ^ (hv >> 23);
	++msgid;
    }
    t.h1 = (h1 ^ (hv << 14)) & 0x7FFFFFFF;	/* bit 31 reserved for future use */
    t.h2 = (h2 ^ (hv << 20)) & 0x7FFFFFFF;	/* bit 31 reserved for future use */
    return(t);
}

char *
MsgId(char *s)
{
    int i;
    static char *LMsgId;

    if (LMsgId) {
	free(LMsgId);
	LMsgId = NULL;
    }

    if (s == NULL)
	return("<>");
    while (*s && (*s == ' ' || *s == '\t'))
	++s;
    if (*s != '<')
	return("<>");
    for (i = 0; s[i] && s[i] != '>'; ++i)
	;
    if (s[i] != '>')
	return("<>");
    if (i > MAXMSGIDLEN)
	return("<>");
    s[i+1] = 0;
    LMsgId = strcpy(malloc(strlen(s) + 1), s);
    return(LMsgId);
}
